/*
Firmware for K106 RGB Laser rev 0.0
Peter Krogen
May 2006
http://krazerlasers.com
Licensed under Creative Commons Attribution-Noncommercial-Share Alike 3.0 United States License
<http://creativecommons.org/licenses/by-nc-sa/3.0/us/>
 which essentially states that you can do whatever you want with it, so long as you
    * attribute me
    * share it under a similar license
    * do not sell it 
If you need a different license, contact me. 
*/


#include <LiquidCrystal.h>
#include <dataflash.h>
#include "WConstants.h" //neededd to make dataflash.h work
#undef int              //neededd to make dataflash.h work
#undef abs              //neededd to make dataflash.h work
#undef double           //neededd to make dataflash.h work
#undef float            //neededd to make dataflash.h work
#undef round            //neededd to make dataflash.h work

/************************ Pin Assignments***************/
// this macro does fast digital write on pins shared with the analog port, (pin 0 is digital pin 14)
#define fastWriteA(_pin_,_state_) (_state_ ?  PORTC |= 1 << (_pin_ ) : PORTC &= ~(1 << (_pin_ ) )) 
//pin assignemnts
#define LCD_RS 2         //lcd rs 
#define LCD_RW 4         //lcd rw 
#define LCD_ENABLE 7     //lcd enable
#define BACKLIGHT 0      //lcd backlight is connected to analog 0
#define LCD_DATA0 8      //lcd data0
#define LCD_DATA1 12     //lcd data1
#define LCD_DATA2 13     //lcd data2
#define LCD_DATA3 3      //lcd data3
#define R_LED 6          //red laser
#define G_LED 5          //green laser
#define B_LED 11         //blue laser
#define X_POS 10          //x galvo
#define Y_POS 9         //y galvo
#define BUTTON_ANALOG 5  //the button is connected to analog 5
#define BUTTON_DIGITAL BUTTON_ANALOG + 14  //when digitalRead is used, analong pin 5 is mapped to 19

//constants
#define BLANK 32         //this is the bytecode for a blank character on the lcd, used to pad empty spaces in the text buffer
#define SELECT 3         //constant returned by readButton for pressing select
#define NONE 0           //constant returned by readButton if no buttons are pressed
#define UP 1             //constant returned by readButton for pressing up
#define DOWN 2           //constant returned by readButton for pressing down
#define BUFFER 1         //the buffer to use on the dataflash (can be 1 or 2)
#define PACKET_LENGTH 8  //the length of each 'packet' stored in the data flash - specifically the packet format is delay,red,green,blue,x(high bit),x(low bit),y(high bit),y(low bit)
#define END 255          //if the the firt bit of a packet is equal to 255, it signals the end of the program (this also means that if you run into unprogramed flash memory, it will signal the end of the program
#define PRESSED 0        //if the select button is pressed, its pin will be high--this used to check that
#define LCDLINES 4       //number if lines available on the LCD
#define MAX_PROGRAMS 63  //maximum number of programs that can be stored in the DataFlash
#define PROGRAM_SIZE 64  //the number of pages in a program
#define READY 0b10101100 //the status returned by the DataFlash when it is ready to be used
#define HEADER_SIZE 24   //the size of the header containing the program name in the date flash

//display modes
#define STANDBY 255        //sit and wait for a user to do something
#define MANUAL 2         //take updates from the serial port and display them directly
#define TEST 0           //display the test circle
#define RANDOM 1         //display random data

/************************Variable Assignments***************/

char text[5][21];


unsigned char r = 0;
unsigned char g = 0;
unsigned char b = 0;
char x = 0;
char y = 0;
int loopDelay = 0;
long a = 12345, m = 10087600, c = 542098091, xrand = 1238764343;

// Connections:
// rs (LCD pin 4) to Arduino pin 2
// rw (LCD pin 5) to Arduino pin 4
// enable (LCD pin 6) to Arduino pin 7
// LCD pin 15 (backlight) to Arduino pin analog 0
// LCD pins d4, d5, d6, d7 to Arduino pins 8, 12, 13, 3
LiquidCrystal lcd(LCD_RS, LCD_RW, LCD_ENABLE, LCD_DATA0, LCD_DATA1, LCD_DATA2, LCD_DATA3);
Dataflash dflash; 


void setup()
{ 
  TCCR0B = 0x01;                          //set the pwm frequiency on timer 0 to 32KHz
  TCCR1B = 0x01;                          //set the pwm frequinecy on timer 1 to 64KHZ
  TCCR2B = 0x01;                          //set the pwm frequinecy on timer 1 to 32KHZ
  clearDisplay();                         //clear the display buffer
  lcd.clear();                            // blank the screen
  DDRC=0x01;                              //set analog 0 (the backlight pin) to an output
  fastWriteA(BACKLIGHT,HIGH);             //turn the backlight on
  randomSeed(19999);   //seed the random generator for random mode
  Serial.begin(9600);                     //open the serial port
  dflash.init();                          //initialize the memory (pins are defined in dataflash.cpp)
  lcd.setCursor(1,0);
  lcd.print("RGV Projector V3.0");
  lcd.setCursor(3,1);
  lcd.print("Peter Krogen");
  lcd.setCursor(6,2);
  lcd.print("ECE 94R");
  lcd.setCursor(3,3);
  lcd.print("Press Select");
  delay(10000);
}

void loop()                            //main loop
{  
 char loop1, loop2;
int i = 0;
unsigned char mode = STANDBY;
char currentProgram = 0;

while(1)
{
 if(digitalRead(BUTTON_DIGITAL) == PRESSED)
  {
    blankLaser();  //turn off the lasers
    clearDisplay();
    for(loop1 = 0; loop1 <  LCDLINES; loop1++) //update the display with the proper program names
      printName(loop1, loop1 + currentProgram);
    printLcd();  //and display them
    while(readButton() != NONE);  //wait until the user lifts their finger off the select button
    
    while(! (readButton() == UP || readButton() == DOWN || readButton() == SELECT));  //and then press a button afterward
    mode = STANDBY;
    while(mode >= MAX_PROGRAMS) 
    {
      if(readButton() == SELECT)
      {
         clearDisplay();
         strcpy(text[0], "currently running:");
         printName(1,currentProgram);
         strcpy(text[2], "press select to stop");  
         mode = currentProgram;  
         printLcd(); 
      }
      else if(readButton() == UP)
      {
         if(currentProgram >= MAX_PROGRAMS)
          currentProgram = 0;
          else
            currentProgram++;
         for(loop1 = 0; loop1 <  LCDLINES; loop1++) //update the display with the proper program names
           printName(loop1, loop1 + currentProgram);
         printLcd();  //and display them
         while(readButton() != NONE);  //wait until the user lifts their finger off the select button
      }
      else if(readButton() == DOWN)
      {
         if(currentProgram == 0) 
           currentProgram = 0;
         else
           currentProgram--;
         for(loop1 = 0; loop1 <  LCDLINES; loop1++) //update the display with the proper program names
           printName(loop1, loop1 + currentProgram);
         printLcd(); 
         while(readButton() != NONE);  //wait until the user lifts their finger off the select button
      }     
    }
      while(readButton() != NONE);  //wait until the user lifts their finger off the select button
  }
   if( mode == TEST )
    projectCircle();
    else if(mode == RANDOM)
   projectRandom();
   else if(mode == MANUAL)
     manualUpdate();
  else if(mode <= MAX_PROGRAMS )
     project(mode);
}
}

//printLcd copies the date stored in the display buffer (text[][]) to the display itself
//it must be called whenever you want to update the display
void printLcd()
{
  char line;
  lcd.clear();                  // start with a blank screen
  for( line = 0; line <= LCDLINES; line++)
  {
    lcd.setCursor(0,line);      // set cursor to column 0, row 0 (the first row)
    lcd.print(text[line]);      // printf the first line of text 
  }
}

//incramentLine moves the text from each of the buffers up one line (ie line 2->1, 3->2, 4->3)
void incramentLine()            
{
  char loop1, loop2;
  for(loop1 = 0; loop1 < LCDLINES - 1; loop1++)
    for(loop2 = 0; loop2 < 20; loop2++)
      text[loop1][loop2] = text[loop1 + 1][loop2];
  return;
}

//ClearLine clears the data stored in display buffer (not the display itself!), specifically the line passed to it.  Call printLcd to update the display
void clearLine(int line)       //fill the buffer given by 'line' with '\0' nulls
{
  char loop1, loop2;
 for(loop1 = 0; loop1 < 19; loop1++);
    text[line][loop1] = BLANK;
    text[line][20] = '\0';
}

//ClearDisplay clears the data stored in display buffer (not the display itself!) call printLcd to update the display
void clearDisplay()       //fill the buffer given by 'line' with '\0' nulls
{
  char loop1, loop2;
  for(loop1 = 0; loop1 < LCDLINES; loop1++)
  {
  for(loop2 = 0; loop2 < 20; loop2++)
   text[loop1][loop2] = BLANK;
  text[loop1][20] = '\0';
  }
}

//DisplayAscii prints a numerical vlaue (of type char) to the LCD, represented in ascii form
//it overwrites any existing data, and leaves the buffer otherwise unnafected
//it puts the most sig bit (padded with 0) at character 'start' on line 'line' 
void displayAscii(char line, char start, unsigned char value)
{
 text[line][start] = ( value / 100) + 48 ;
 value = value - 100*(value / 100);
 text[line][start + 1] = ( value / 10 ) + 48;
 value = value - 10*(value/10);
 text[line][start + 2] = value + 48;
}

char readButton()
{
  
  int val = 0, loop1 = 0;
  for(loop1 = 0; loop1 < 100; loop1++)
   {
     val = val + analogRead(BUTTON_ANALOG) / 10;    // read the value from the sensor
   }
   val = val / 1000;
   if(val == 10)
     return NONE;
   if(val > 7)
     return UP;
   if(val > 5)
      return DOWN;
   else
     return SELECT;
}

void manualUpdate(void)
{
  unsigned  char buffer[6] = {0, 0, 0, 0};
  char character = 0, posisition = 0;
  int pointer = 0;
  byte inByte = 0;
   if (Serial.available() >0)
   {          // read the incoming byte:
    inByte = Serial.read();
    if (inByte == '#') // If the # sign is detected, the data should follow
    {
      while (pointer < 5)
      { // accumulate 6 chars
      if(Serial.available() > 0){
        buffer[pointer] = Serial.read(); // store in the buffer
        pointer++; // move the pointer forward by 1
      }
     }
   // now we have the 3 numbers stored as an unsigned char (0-255)
   r = buffer[0];
   g = buffer[1];
   b = buffer[2];
   x = buffer[3];
   y = buffer[4];    
   displayAscii(0, 0, r);  //copy them into the display buffer
   displayAscii(1, 0, g);
   displayAscii(2, 0, b);
   displayAscii(3, 0, x);
   displayAscii(3, 5, y);
   r = 255 - r;            //and scale them
   g = 255 - (g/2 + 128);
   if( g >= 127) g = 255;
   //g = 255 - g;
   b = 255 - b;
   analogWrite(R_LED, r);  //finally update the output values
   analogWrite(G_LED, g); 
   analogWrite(B_LED, b);
   analogWrite(X_POS, x);
   analogWrite(Y_POS, y);
   pointer = 0; // reset the pointer so we can reuse the buffer
   printLcd();   //update the display
  }
}
}

void blankLaser()
{
   analogWrite(R_LED, 255);  //finally update the output values
   analogWrite(G_LED, 255); 
   analogWrite(B_LED, 255);
   analogWrite(X_POS, 128);
   analogWrite(Y_POS, 128);
}

void projectRandom()
{
  char x_pos, y_pos;
 r = betterRandom();
 g = betterRandom();
 b = betterRandom();
 x_pos = betterRandom();
 y_pos = betterRandom();
 loopDelay = betterRandom();
 r = r / 2;
 b = b / 2;
 r = 255 - r;
 g = 255 - (g/2 + 128);
 if( g >= 127) g = 255;
  b = 255 - b;
 analogWrite(R_LED, r); 
 analogWrite(G_LED, g);
 analogWrite(B_LED, b);
 analogWrite(X_POS, 128 + x_pos / 3 );
 analogWrite(Y_POS, 128 + y_pos / 3);
 delayMicroseconds(loopDelay * 100);
}

long betterRandom()
{
 xrand = ((a * xrand-1) + c-1) % m;
  c = ((a * xrand-1) + c-1) / m;
 return xrand;
}

void projectCircle()
{
  char i;
  char x_pos[102] = {0,6,12,18,24,30,36,42,48,53,58,63,68,72,77,80,84,87,90,92,95,96,98,99,99,100,99,99,98,96,95,92,90,87,84,80,77,72,68,63,58,53,48,42,36,30,24,18,12,6,0,-6,-12,-18,-24,-30,-36,-42,-48,-53,-58,-63,-68,-77,-77,-80,-84,-87,-90,-92,-95,-96,-98,-99,-99,-100,-99,-99,-98,-96,-95,-92,-90,-87,-84,-80,-77,-72,-68,-63,-58,-53,-48,-42,-36,-30,-24,-18,-12,-6,0};
  char y_pos[102] = {100,99,99,98,96,95,92,90,87,84,80,77,72,68,63,58,53,48,42,36,30,24,18,12,6,0,-6,-12,-18,-24,-30,-36,-42,-48,-53,-58,-63,-68,-72,-77,-80,-84,-87,-90,-92,-95,-96,-98,-99,-99,-100,-99,-99.,-98,-96,-95,-92,-90,-87,-84,-80,-77,-72,-68,-63,-58,-53,-48,-42,-36,-30,-24,-18,-12,-6,0,6,12,18,24,30,36,42,48,53,58,63,68,72,77,80,84,87,90,92,95,96,98,99,99,100};
  for(i = 0; i <=  100; i = i + 1)
  {
    x = 128 + ((int)x_pos[i] / 3);
    y = 128 + ((int)y_pos[i] / 3) ;
    if(i < 20 )
     {
       r = 0;
       g = 255;
       b = 0;
     }
     else if(i < 30)
     {
       r = 0;
       g = 255;
       b = 120;
     }
     else if(i < 40)
     {
       r = 60;
       g = 255;
       b = 0;
     }
     else if(i < 50 )
     {
       r = 130;
       g = 255;
       b = 0;
     }
     else if(i < 60 )
     {
       r = 250;
       g = 0;
       b = 0;
     }
     else if(i < 70 )
     {
       r = 37;
       g = 0;
       b = 250;
     }
     else if(i < 80 )
     {
       r = 0;
       g = 0;
       b = 255;
     }
     else
     {
       r = 0;
       g = 255;
       b = 120;
     }
    r = 255 - r;
    g = 255 - (g/2 + 128);
    if( g >= 127) g = 255;
    b = 255 - b;
    analogWrite(R_LED, r);  
    analogWrite(G_LED, g); 
    analogWrite(B_LED, b);
    analogWrite(X_POS, x);
    analogWrite(Y_POS, y);
  }
}

void project( char program )
{
    while( dflash.Read_DF_status() != READY);
      dflash.Cont_Flash_Read_Enable(program * PROGRAM_SIZE, HEADER_SIZE);
      loopDelay = dflash.DF_SPI_RW(0);
      while(loopDelay != END) //32 depends on the amount of data on the page
      {
            r = dflash.DF_SPI_RW(0);
            g = dflash.DF_SPI_RW(0); 
            b = dflash.DF_SPI_RW(0);
            x = dflash.DF_SPI_RW(0) * 255;
            x = x + dflash.DF_SPI_RW(0);
            y = dflash.DF_SPI_RW(0) * 255;
            y = y + dflash.DF_SPI_RW(0);
     /*       
         Serial.print(loop_delay, DEC);
         Serial.print('\t');
         Serial.print(r, DEC);
         Serial.print('\t');
         Serial.print(g, DEC);
         Serial.print('\t');
         Serial.print(b, DEC);
         Serial.print('\t');
         Serial.print(x, DEC);
         Serial.print('\t');
         Serial.print(y, DEC);
         Serial.print('\t');
         Serial.println();
       */     
            r = 255 - r;
            g = 255 - (g/2 + 128);
            if( g >= 127) g = 255;
            b = 255 - b;
            delayMicroseconds(100 * loopDelay);
            analogWrite(R_LED, r); 
            analogWrite(G_LED, g);
            analogWrite(B_LED, b);
            analogWrite(X_POS, x);
            analogWrite(Y_POS, y);
            loopDelay = dflash.DF_SPI_RW(0);
      }
      dflash.DF_CS_inactive(); 
}

void printName( char line, char program )
{
  char loop1 = 0, character;
  clearLine(line);
  if(program < MAX_PROGRAMS)
  {
    while( dflash.Read_DF_status() != READY);
      dflash.Cont_Flash_Read_Enable(program * PROGRAM_SIZE, 0);
      for(loop1 = 0; loop1 < 20; loop1++)
      {
           character = dflash.DF_SPI_RW(0);
           if(character == '\0')
               text[line][loop1] = BLANK;
            else
              text[line][loop1] = character;
      }
      dflash.DF_CS_inactive(); 
  }
}
